unit Snake;

// Author Elias Konstadinidis
// email : coma@onf.com

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, extctrls;

type
  TDirection = (drUp, drRight, drDown, drLeft);
  TSnake = class(TWinControl)
  private
    { Private declarations }
    FOnGameOver : TNotifyEvent;
    FOnScoreUpdate : TNotifyEvent;
    procedure SnakeTimer(Sender: TObject);
    function GetSpeed: integer;
    procedure SetSpeed(const Value: integer);
  protected
    { Protected declarations }
    FScore : integer;
    ImageBuffer : TImage;
    snake : record
      Tail : Array[0..255] of TPoint;
      Length : integer;
      Direction : TDirection;
    end;
    Timer : TTimer;
    Dot : TPoint;
    function DotInTail(vDot : TPoint) : boolean;
    procedure MoveSnake; virtual;
    procedure PrintSnake; virtual;
    procedure InitGame;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure Resize; override;
    procedure ConstrainedResize(var MinWidth, MinHeight, MaxWidth,
      MaxHeight: Integer); override;
    procedure SetParent(AParent: TWinControl); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property Speed : integer read GetSpeed write SetSpeed;
    property Score : integer read FScore;
    property OnGameOver : TNotifyEvent read FOnGameOver write FOnGameOver;
    property OnScoreUpdate : TNotifyEvent read FOnScoreUpdate write FOnScoreUpdate;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Game Controls', [TSnake]);
end;

{ TSnake }

procedure TSnake.ConstrainedResize(var MinWidth, MinHeight, MaxWidth,
  MaxHeight: Integer);
begin
  inherited;
  MinWidth := 32; MinHeight := 20;
end;

constructor TSnake.Create(AOwner: TComponent);
begin
  inherited;
  randomize;
  Width  := 320;
  Height := 200;
  TabStop := true;
  ImageBuffer := TImage.Create(self);
  ImageBuffer.Name   := 'SnakeImage';
  ImageBuffer.Width  := 320;
  ImageBuffer.Height := 200;
  ImageBuffer.Top  := 0;
  ImageBuffer.Left := 0;
  ImageBuffer.Picture.Bitmap.Width  := 320;
  ImageBuffer.Picture.Bitmap.Height := 200;
  ImageBuffer.Stretch := true;
  ImageBuffer.Parent := self;
  Timer := TTimer.Create(self);
  Timer.Name     := 'SnakeTimer';
  Timer.OnTimer  := SnakeTimer;
  Timer.Interval := 400;


  FOnGameOver    := nil;
  FOnScoreUpdate := nil;

  InitGame;
end;

destructor TSnake.Destroy;
begin
  if Assigned(Timer) then
    FreeAndNil(Timer);
  if Assigned(ImageBuffer) then
    FreeAndNil(ImageBuffer);
  inherited;
end;

function TSnake.DotInTail(vDot: TPoint): boolean;
var
  i : integer;
begin
  result := false;
  for i := 0 to snake.Length-1 do
    if (snake.Tail[i].x = vDot.x) and
       (snake.Tail[i].y = vDot.y) then
      result := true;
end;

function TSnake.GetSpeed: integer;
begin
  result := Timer.Interval;
end;

procedure TSnake.InitGame;
begin
  if csDesigning in ComponentState then
    Timer.Enabled := false
  else
    Timer.Enabled := true;

  snake.Length := 3;
  snake.Tail[0].x := 10;
  snake.Tail[0].y := 10;
  snake.Tail[1].x := 11;
  snake.Tail[1].y := 10;
  snake.Tail[2].x := 12;
  snake.Tail[2].y := 10;
  snake.Direction := drRight;

  Dot.x := -1;
  Dot.y := -1;

  FScore := 0;
  if Assigned(FOnScoreUpdate) then
    FOnScoreUpdate(self);
end;

procedure TSnake.KeyUp(var Key: Word; Shift: TShiftState);
begin
  inherited;
  case Key of
    VK_UP    : snake.Direction := drUp;
    VK_DOWN  : snake.Direction := drDown;
    VK_LEFT  : snake.Direction := drLeft;
    VK_RIGHT : snake.Direction := drRight;
  end;
end;

procedure TSnake.MoveSnake;
var
  i : integer;
  newPoint : TPoint;
begin
  // put new dot
  if Dot.x = -1 then begin
    repeat
      Dot.x := random( 32 );
      Dot.y := random( 20 );
    until not DotInTail(Dot);
  end;

  newPoint := snake.Tail[snake.Length-1];
  case snake.Direction of
    drUp    : dec(newPoint.y);
    drRight : inc(newPoint.x);
    drDown  : inc(newPoint.y);
    drLeft  : dec(newPoint.x);
  end;
  if (newPoint.x < 0) or
     (newPoint.x = 32) or
     (newPoint.y < 0) or
     (newPoint.y =20) or
     DotInTail(newPoint) then begin
    Timer.Enabled := false;
    if Assigned(FOnGameOver) then
      FOnGameOver(self);
    InitGame;
    exit;
  end;
  if (Dot.x = newPoint.x) and (Dot.y = newPoint.y) then begin
    //eat dot
    Dot.x := -1;
    Dot.y := -1;
    inc(FScore);
    if Assigned(FOnScoreUpdate) then
      FOnScoreUpdate(self);
    if snake.Length < 255 then begin
      inc(snake.Length);
      snake.Tail[snake.Length-1] := newPoint;
    end;
  end else begin
    for i:= 1 to snake.Length-1 do
      snake.Tail[i-1] := snake.Tail[i];
    snake.Tail[snake.Length-1] := newPoint;
  end;
end;

procedure TSnake.PrintSnake;
var
  dRect : TRect;
  i : integer;
begin
  if not Assigned(ImageBuffer) then
    Exception.Create('ImageBuffer has been released');
  with ImageBuffer.Picture.Bitmap do begin
    Canvas.Brush.Color := clBlack;
    Canvas.Pen.Color   := clLime;
    Canvas.Rectangle(Canvas.ClipRect);

    Canvas.Brush.Color := clGreen;
    for i := 0 to snake.Length-1 do begin
      SetRect(dRect, snake.Tail[i].x*10, snake.Tail[i].y*10, snake.Tail[i].x*10+10, snake.Tail[i].y*10+10);
      Canvas.FillRect(dRect);
    end;
    // print dot
    Canvas.Brush.Color := clLime;
    if Dot.x <> -1 then begin
      SetRect(dRect, Dot.x*10, Dot.y*10, Dot.x*10+10, Dot.y*10+10);
      Canvas.FillRect(dRect);
    end;
  end;
end;

procedure TSnake.Resize;
begin
  inherited;
  ImageBuffer.Width  := Width;
  ImageBuffer.Height := Height;
end;

procedure TSnake.SetParent(AParent: TWinControl);
begin
  inherited;
  if (csDesigning in ComponentState) and
     (not (csDestroying in ComponentState)) then
    PrintSnake;
end;

procedure TSnake.SetSpeed(const Value: integer);
begin
  Timer.Interval := Value;
end;

procedure TSnake.SnakeTimer(Sender: TObject);
var
  BitmapChange : TNotifyEvent;
begin
  MoveSnake;
  BitmapChange := ImageBuffer.Picture.Bitmap.OnChange;
  ImageBuffer.Picture.Bitmap.OnChange := nil;
  PrintSnake;
  ImageBuffer.Picture.Bitmap.OnChange := BitmapChange;
  BitmapChange(ImageBuffer.Picture.Bitmap);
end;

end.
